/*----------------------------------------------------------------------
    kiri  -  cut over-connection of text line   ver. 2.1
        Created  by H.TSUKATA     1994
        Modified by H.Goto    Sep 1997
        Modified by H.Goto    Jul 2003
----------------------------------------------------------------------*/

/*--------------------------------------------------------------------
  Copyright (C) 1994-2003  Hitoshi TSUKATA and Hideaki GOTO

        All Rights Reserved

  Permission to use, copy, modify, and distribute this software and
  its documentation for any purpose is hereby granted without fee,
  provided that (i) the above copyright notice and this permission
  notice appear in all copies and in supporting documentation, (ii)
  the authors' names, Hitoshi TSUKATA and Hideaki GOTO, may not be
  used in any advertising or otherwise to promote the sale, use or
  other dealings in this software without prior written authorization
  from the authors, (iii) this software may not be used for
  commercial products without prior written permission from the
  author, and (iv) the notice of modification is specified in cases
  where modified copies of this software are distributed.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND.
  THE AUTHOR WILL NOT BE RESPONSIBLE FOR ANY DAMAGE CAUSED BY THIS
  SOFTWARE.
--------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "seg.h"
#include "line.h"

// for C++ STL
#include <list>

#define TITLE		"kiri   ver. 2.1, Copyright (C) 1994-2003 H.Tsukata and H.Goto "

OPTION	opt;	

/*----------------------------------------------
  File output functions
----------------------------------------------*/

void fo_exit( char *fn){
  fprintf(stderr,"File error , Can't Open %s\n", fn);
  exit(1);
}

FILE *fo( char *fn, char *mode)
{
  FILE	*fptmp;
  
  if( fn==NULL) fo_exit( fn);
  if( strlen( fn)<1){
	// if filename is omitted ...
	if( strchr( mode, 'w'))	return( stdout);
	else	return( stdin);
  }  else{
	if( (fptmp=fopen( fn, mode)) ==NULL){
	  fo_exit( fn);
	} else{
	  fclose( fptmp);
	  return( fopen( fn, mode));
	}
  }
  return NULL;
}

/*--------------------------------------------*/


void printusage(void) {
  fprintf( stderr,"%s\n", TITLE);
  fprintf( stderr,"Usage : %% kiri <String FileName> -limgdir <Line Image Directory> [-o <Output File>] [-dpi res (default 200)]\n");
  exit(1);
}
	
void getargv(int argc, char **argv){
  int i=1,errflg=0;
  strcpy( opt.com,argv[0]);

  /***** ǥե *****/
  opt.haba = HABA_DEFAULT;
  opt.high = HIGH_DEFAULT;
  strcpy( opt.lineimagedir, LINEIMAGEDIR_DEFAULT );


  /*****  *****/
  for( i=1; i<argc; i++){
	if ( argv[i][0] !=  '-'){
	  /*---------------- ϥե̵ ----------------*/
	  if ( errflg != 0){   printusage();	  }
	  strcpy( opt.strfile, argv[i]);
	  errflg = 1;
	} else{
	  /*---------------- ϥե󤢤 ----------------*/
	  /* ץΤ */
	  if ( strstr( argv[i], "verbose") != NULL)		opt.msg_flg = 1;
	  else if ( strstr( argv[i], "info") != NULL)	opt.info = 1;
//	  else if ( strstr( argv[i], "debug") != NULL)	opt.debug_flg = 1;
	  else if ( strstr( argv[i], "v") != NULL)		opt.msg_flg = 1;
	  else {
		/* ץ+ */
		if( i+1 >= argc){ printusage(); } 
		if ( strstr( argv[i], "limgdir") != NULL ){
		  strcpy( opt.lineimagedir, argv[++i]);
		  strcat( opt.lineimagedir, "/");
		}else if ( strstr( argv[i], "o") != NULL )
		  strcpy( opt.outfile, argv[++i]);
		else if(strstr( argv[i], "ha") != NULL)
		  opt.haba = atoi( argv[++i]);
		else if(strstr( argv[i], "hi") != NULL)
		  opt.high = atoi( argv[++i]);
		else if(strstr( argv[i], "dpi") != NULL)
		  opt.dpi = atoi( argv[++i]);
		else {
		  fprintf( stderr, "%s Argument Error !\n", argv[0]);
		  printusage();
		}
	  }
	}
  }

  if( errflg == 0){
	printusage();
	exit(1);
  }

#if 0
  if( opt.msg_flg){
	fprintf( stderr, "Thresholds are the following :\n");
	fprintf( stderr, "\t haba : %d \n", opt.haba);
	fprintf( stderr, "\t high : %d \n", opt.high);
  }
#endif

  if( 	strlen(opt.outfile)<1){
	strcpy( opt.outfile, opt.strfile);
	strcat( opt.outfile, ".kiri");
  }



}


void rls(int *in, int **out, int c, int flg){
   /*------------ Ran Length Smoothing -------------- */
   int	k,y,i;

   for( k=0; in[k]>=0; k++);
   (*out) = (int *)calloc( k+1, sizeof( int));

   i=0;
   while(  *in >= 0 ){	//οä餪.
      if( (*in==0)*(flg==1) + (*in>0)*(flg==0) ){
	 k=0;
	 while( ((*in==0)*(flg==1) + (*in>0)*(flg==0)) && *in >= 0){
	    in++;
	    k++;
	 }
	 y = (k<c);
	 for( ; k>0 ; k--, i++){
	    (*out)[i] = y*(flg==1) + (1-y)*(flg==0); ;
	 }
      } else {
	 (*out)[i] = (flg==1);
	 i++;
	 in++;
      }
   }
   
   (*out)[i] = -1;
}


void mabiki(int *prj, int c){
   /* projection Τ 礭 c ̤ ΤΤ 0 ˤ */
   while( *prj>=0){
      *prj = *prj * ( *prj >= c);
      prj++;
   }
}

int * makelineprojection(STRING *sp)
{
   /*	node ʸԥ᡼Υҥȥ prj ˳Ǽ.
    *	
    *
    */
   char		imgfn[512];
   int		i;
   int		*prj;
   pbm_file	img;
   
   sprintf( imgfn, "%sSTR%04d", opt.lineimagedir, sp->index);
   img.readpbm(imgfn); 
   if( img.pbmprojection( &prj, 0)<0){
	 fprintf( stderr, "%s: pbm dataerr in %s \n", opt.com, imgfn);
	 exit(1);
   }
   if( opt.dpi==400){		/*======= 400dpi ΤȤ =======*/
	 // *** 400->200 ؤΡְ̡
	 for(i=0 ; prj[i]>=0; i++){
	   prj[i/2]=(int)(prj[i]/2.0);	/* header.pnm_width  ɬ */
	 }
	 prj[i/2]=-1;      
   }
   return prj;

}


void makerunlength(int *projection, STRING *sp){
   int	i=-1,j=0,k=0;
   int	flg, oldflg = RUN_TOP;
   int	*conved_prj;

/*
   mabiki( projection, 3);
*/
   rls( projection, &conved_prj , (int)( sp->h * T100) , 0);
   
   j=0;
   while( conved_prj[k] >= 0){
	 flg = ( conved_prj[k] !=  0);
	 if ( flg != oldflg ){
	   sp->run[j] = i+1;
	   j++;
	   i = 0;
	 } else {
	   i++;
	 }
	 oldflg = flg;
	 k++;
   }
   sp->run[j]= i+1;
   j++;
   sp->run[j]= -1;
}


void makecuttmp (STRING *line){
  /*
   *	STRING line->run   ǸͱƥǼ
   */

  int	width;
  int	*projection;
  STRING	*sp;

  for ( sp=line; sp->chk_marker()==0; sp++){
	projection = makelineprojection( sp);
	for( width=0; projection[width]>=0; width++);
	sp->run = (int *)calloc( width, sizeof(int));
	makerunlength( projection, sp);
	sp->makeccand();
	free( projection);
  }

   if(opt.msg_flg)
	 fprintf( stderr, "  Input  %3d Lines\n", (sp-1)->index+1);
   
}


void smoothing(int *cutproj, int h){
  int	i,j,len ;
  int	*buf;
  
  for( len=0 ; cutproj[len]<INT_MAX ; len++);
  buf = (int *) calloc( len+h*3, sizeof(int));
  
  /*
  for( i=h ; cutproj[i+h]<INT_MAX ; i++){
	buf[i] = 0;
	for( j=-h ; j<=h ; j++){
	  buf[i] += cutproj[i+j];
	}
  }
  */
  for( i=0; i<len ; i++){
	buf[i] = 0;
	for( j=((i-h>0)? i-h:0); j<((i+h<len)? i+h:len);  j++)
	  buf[i] += cutproj[j];
  }
  
  for( i=0; i<len; i++)	cutproj[i] = buf[i];
  
  free(buf);
}


std::list<int> *searchpeak( STRING *sp, int *cutproj){
  /*
   *	** ԡФƥꥹȤ֤ ( **кɸ**)
   */
  static int	oldsflg,sflg  ;
  static int	oldnode, oldpeak ;
  int		xl, pflg, k; 
  int		peak; // (кɸ)
  int		rx;
  int		ip;
  std::list<int>	*peaks;
  int		itmp;

  peaks = new std::list<int>;
  for( ip=0; ip < sp->w; ){
	
	/* Τܤꤶ || ꤶ */
	oldsflg = sflg;
	sflg = seifu( cutproj[ip+1] - cutproj[ip] ) ;
	ip++;

	if( oldsflg==1 && sflg!=1){
	  xl = ip;
	  while ( ip<sp->w-1 && sflg==0 ){
		/* ؤ */
	    oldsflg = sflg;
	    sflg = seifu( cutproj[ip+1] - cutproj[ip] ) ;
	    ip++;
	  }

	  if ( sflg==-1 && cutproj[(xl+ip-1)/2] >= opt.high ){
		/* ؤ   */
	    peak = ( ip + xl ) / 2;  
		
		/*   å */
	    pflg =0;	//åѥե饰
	    for( k=0,rx=sp->run[0]; sp->run[k]>=0 ; k++,rx+=sp->run[k]){
		  if( sp->run[k]> sp->h*T72 
			  && rx>peak  ){
			if ( (k+RUN_TOP)%2 == 0
				 && rx-sp->run[k]<=peak){
			  pflg = 1 ;
			  break;
			} 
		  }
	    }

	    if( pflg == 1){
		  if ( peak > SLIT_WIDTH && peak < sp->w-SLIT_WIDTH ){ 
			if( !( oldnode==sp->index && oldpeak==peak)){
			  oldnode = sp->index;
			  oldpeak = peak;
			  itmp = peak+sp->x;// кɸкɸ
			  peaks->push_back( itmp);	// ** ɸ㡼*
			}
		  }
	    } else 
		  if( opt.info)
			fprintf( stdout,"%4d(-) ", peak); 
	  }
	}
  }

  return (peaks);
}


void STRING::makeccand(void){
  /* Ǹõ 
   *        -> (STRING::) c_cand[0|1] ˳Ǽ
   */ 
  int	j, rx;
  
  // ü
  c_cand[0].clear();
  c_cand[0].push_back( -16);
  c_cand[0].push_back( w+16);
  
  // 
  c_cand[1].clear();
  for( j=0,rx=run[0] ; run[j]>=0 ; j++, rx+=run[j]){
	if(  run[j] > h*T71  && (j+RUN_TOP)%2 == 0 ) {
	  c_cand[1].push_back( (int)( rx - run[j]/2.0));
	}
  }

}

      

int *cut_prj( STRING *sp){
  /*---------------- Ǹ ͱ 롼----------------*/
  int		i;
  double	px;	/* кɸ */
  double	rx;	/* кɸ */    
  double	dx;
  std::list<STRING *>	sl;
  STRING	*sp1;
  
  int		*cutproj; // ͱѥ졼
  cutproj = (int *)calloc( sp->w+1, sizeof( int));  /*  */

  /* ˵ʸԽFSͱƤ */
  
  for( sl=sp->adj; ! sl.empty() ; sl.pop_front() ){
	sp1=sl.front();
	//dx = sin( deg2rad( sp->slope+sp1->slope)/(double)2 )
	// linedistance( sp, sp1);
	dx = 0;
	std::list<int> cl;
	for( i=0; i<2; i++){
	  for( cl=sp1->c_cand[i]; ! cl.empty() ; cl.pop_front()){
		rx=cl.front();
		px =  sp1->x + rx - dx ;
		if ( px>sp->x && px<sp->x+sp->w ){
		  cutproj[ (int)px - sp->x] += ( 1 + ( i==1));
		  //   ֥󥯤 2 Ф򤫤Ƥ ͱ 
		}
	  }
	}
  }
  cutproj[ sp->w] = INT_MAX ; 
  
  return cutproj;
}



void makecut(STRING *line ){
  int		*cutproj;
  FILE 	*outfp;
  STRING *sp;
  int	cut_cnt, line_cnt;

/*
  FILE *fp, *fp0, *fp1;

  if( opt.debug_flg){
	fp = fo( "./kiritmp-ps/fs_cand.ps", "w");
	fp0 = fo( "./kiritmp-ps/fs_cand0.ps", "w");
	fp1 = fo( "./kiritmp-ps/fs_cand1.ps", "w");
	printpsheader( fp);
	printpsheader( fp0);
	printpsheader( fp1);
  }
*/

  for ( sp=line,cut_cnt=0,line_cnt=0; sp->chk_marker()==0;
 		sp++,line_cnt++){ // ʸԤˤĤ...
	sp->cutpoints = new std::list<int>;
	if(opt.info)  fprintf( stdout,"\nString #%3d :    ", sp->index);
	/***** ͱ *****/
	cutproj = cut_prj( sp);
	
	/***** ͱƥҥȥ ʿ경 *****/
	smoothing( cutproj, 3); 
	smoothing( cutproj, opt.haba);

	/***** ԡ *****/
	sp->cutpoints = searchpeak( sp, cutproj);	
	cut_cnt += sp->cutpoints->size();

	if( opt.debug_flg){/* Debug */
	  char	fn[512];
	  int	k;
	  sprintf( fn, "./kiritmp/K%04d.dat", sp->index);
	  FILE *fpd = fo( fn, "w");
	  for( k=0; cutproj[k]<INT_MAX; k++)
	    fprintf( fpd, "%d %d\n", k, cutproj[k]) ;
	  for( std::list<int> l=*(sp->cutpoints); ! l.empty() ; l.pop_front())
	    fprintf( fpd, "%d -1\n", l.front()-sp->x) ;
	  fclose(fpd);
	}/* Debug */

#if 0
	if( opt.debug_flg){
	  std::list<int> cl;
	  for( int i=0; i<2; i++){
		for( cl=sp->c_cand[i]; ! cl.empty() ; cl.pop_front()){
		  fprintf( fp, "%d %d 15 0 360 arc ", sp->x+cl.front(), sp->y);
		  fprintf( fp, "stroke %s\n", (i==0)? "" : "fill");
		  fprintf( (i==0)?fp0:fp1, "%d %d 15 0 360 arc fill\n", 
				  sp->x+cl.front(), sp->y);
		}
	  }
	}
#endif
	
  }

#if 0
  if( opt.debug_flg){
	fclose( fp);
	fclose( fp0);
	fclose( fp1);
  }
#endif
  
  // ʸ
  Document	*new_doc;
  STRING	*new_sp;
  std::list<int>	peaks;	
  int		peak;
  Rectangle	*bp, *new_bp;
  Rectangle 	*btmp = 0;

  short	mid_cut_flg=0;  //ľˡPR ˤǡפäɤ.
  int	cut_cnt_rel=0;
  new_doc = new Document();
  new_doc->line = new STRING[ line_cnt+cut_cnt+1];
  for( sp=line, new_sp=new_doc->line; sp->chk_marker()==0; sp++){
	bp=sp->box;
	if( sp->cutpoints->size()>0){
	  for( peaks=*(sp->cutpoints); ! peaks.empty() ; peaks.pop_front()){// ˤĤ...
		peak=peaks.front() ;			// peak; **ɸ
		new_sp->box = new Rectangle[ sp->box_num+2];	//ʸԤζΰ

		if ( opt.info)	fprintf( stdout,"%4d ",peak );// * Debug *
	

		// ޤǤΥǡ򥳥ԡ
		for( new_bp=new_sp->box; bp->x+bp->w <= peak; new_bp++,bp++ ){
		  if( mid_cut_flg){
			*new_bp= *btmp; delete btmp;
			mid_cut_flg = 0;
		  } else{
			*new_bp = *bp ;
		  }
		}
		// ^^^^^!!!!!!!!!!!!! ǽ !!!!!!!!!!^^^^^^
		if(  peak > bp->x ){
		  // PR ˤ 
		  btmp = new Rectangle( bp->x, bp->y, peak-bp->x, bp->h) ;
		  *(new_bp++)= *btmp;	delete btmp;
		  (new_bp++)->set_marker( 0);
		  mid_cut_flg = 1;
		  btmp = new Rectangle( peak, bp->y, bp->x+bp->w-peak, bp->h) ;
		} else {
		  // PR ˤ 
		  (new_bp++)->set_marker(0);
		}
		if( new_sp->box[0].chk_marker() ==0){
		  (new_sp++)->calc_data() ;
		  cut_cnt_rel++;
		} else{
		  delete(new_sp->box);
		  // ˤΤ ɤʤȤǤ礦
		  //      ԡܶᤷƤƤΥԡδ֤ˤҤȤĤζ
		  //      ʤ硥
		  //			   C C C C p p C C C C
		  //			...R R R R 1 2 R R R R...
		  

		}
	  }
	} 
	  
	// ⤦ԡ ̵.
	new_sp->box=new Rectangle[ sp->box_num+2];
	new_bp = new_sp->box;
	do{
	  if( mid_cut_flg){
		*new_bp= *btmp;	delete btmp;
		mid_cut_flg = 0;
	  } else{
		*new_bp = *bp ;
	  }
	  new_bp++; bp++;
	} while ( bp->chk_marker()==0);

	(new_sp++)->calc_data() ;
	cut_cnt_rel++;

  }

  new_doc->writelinedata( outfp=fo( opt.outfile, "wb"));
  fclose(outfp);
   
  if( opt.msg_flg)	fprintf( stderr,"  Output %3d Lines\n"
							 ,cut_cnt_rel);
  if( opt.info) 	fprintf( stdout,"\n"); 
      
}


int main (int argc, char **argv){

  FILE		*fp;
  Document	*doc;

  doc = new Document();
  getargv( argc, argv);
  if( opt.msg_flg)  fprintf( stderr, "-- %s --\n", TITLE);

  doc-> readlinedata( fp=fo( opt.strfile, "rb") );
  
   
  makenss( doc->line);
  makecuttmp( doc->line);
  makecut( doc->line);
   
  return(0);

}
