/*------------------------------------------------------
	Library for Document Analysis System
	(Document Image Frame-buffer class)
		Written  by H.Goto, Feb. 1995
		Modified by H.Goto, Feb. 1997
		E-Mail: hgot@ecip.tohoku.ac.jp
------------------------------------------------------*/

/*--------------------------------------------------------------------
  Copyright (C) 1995-1997  Hideaki Goto

        All Rights Reserved

  Permission to use, copy, modify, and distribute this software and
  its documentation for any purpose is hereby granted without fee,
  provided that (i) the above copyright notice and this permission
  notice appear in all copies and in supporting documentation, (ii)
  the name of the author, Hideaki Goto, may not be used in any
  advertising or otherwise to promote the sale, use or other
  dealings in this software without prior written authorization
  from the author, (iii) this software may not be used for
  commercial products without prior written permission from the
  author, and (iv) the notice of the modification is specified in
  case of that the modified copies of this software are distributed.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND.
  THE AUTHOR WILL NOT BE RESPONSIBLE FOR ANY DAMAGE CAUSED BY THIS
  SOFTWARE.
--------------------------------------------------------------------*/


#include	<stdio.h>
#include	<stdlib.h>

#include	"Docfb.h"



DOCFB :: DOCFB(){
	image = 0;
	bgcolor = 0;
	xpad = ypad = 1;
}


DOCFB :: ~DOCFB(){
	if ( image != 0 )  sip_DestroyImage(image);
}


void DOCFB :: setpadding(int xpad,int ypad){
	DOCFB::xpad = xpad;
	DOCFB::ypad = ypad;
}


char * DOCFB :: getimgptr(int x,int y){
	if ( image == 0 )  return(0);
	if ( image->depth == 1 )  x /= 8;
	return((char *)sip_getimgptr(image,y) + x);
}


int DOCFB :: load_file(char *fname,int noload,int transpose){
					/* ret = -3 : Illegal file type */
					/*       -2 : File read error   */
					/*       -1 : Memory over       */
	PBMFILE		PBMLD;
	int		y;
	int		errflag;
	char		*line;
	int		depth;
	int		width8,fwidth,fheight;
	errflag = PBMLD.open(fname,"r");
	switch ( errflag ){
	  case 0:	break;
	  case -5:	return(-3);
	  case -6:	return(-1);
	  default:	return(-2);
	}
	switch( PBMLD.getfiletype() ){
	  case 0:	depth = 1;
			break;
	  case 1:	depth = 8;
			break;
	  default:	return(-3);
	}
	if ( ! transpose ){
		width0 = fwidth = PBMLD.getwidth();
		height0 = fheight = PBMLD.getheight();
	}
	else{
		width0 = fheight = PBMLD.getheight();
		height0 = fwidth = PBMLD.getwidth();
	}
	width  = ((width0  + xpad -1) / xpad) * xpad;
	height = ((height0 + ypad -1) / ypad) * ypad;
	if ( depth == 1 )  width8 = ((fwidth + 7) /8) *8;
	else               width8 = fwidth;
	if ( image != 0 )  sip_DestroyImage(image);
	if ( 0 == (image = sip_CreateImage(width,height,depth)) )  return(-1);
	if ( 0 == (line = new char[fwidth]) )  return(-1);
	errflag = 0;
	if ( ! noload ){
		if ( depth == 1 )  sip_ClearImage(image,(long)(0xff - bgcolor));
		else               sip_ClearImage(image,(long)bgcolor);
		for ( y=0 ; y<fheight ; y++ ){
			if ( 0 != PBMLD.readline(-1,line) ){
				errflag = 1;  break;
			}
			if ( ! transpose ){
				sip_PutScanLine(image,(void *)line,0,y,width8,depth);
			}
			else{	sip_PutVertLine(image,(void *)line,y,0,width8,depth);
			}
		}
	}
	delete line;
	PBMLD.close();
	if ( errflag )  return(-2);
	return(0);
}



