/*--------------------------------------------------------------
	String Extraction from Image   strextr  v2.4
		Written  by H.Goto , Dec 1994 (original v1.0)
		Modified by H.Goto , Nov 1995
		Modified by H.Goto , Jul 1996
		Modified by H.Goto , Aug 1996
		Modified by H.Goto , Oct 1996
		Modified by H.Goto , Nov 1996
		Modified by H.Goto , Dec 1996
		Modified by H.Goto , Aug 1997 (revised for O2)
		Modified by H.Goto , Jul 2003
--------------------------------------------------------------*/

/*--------------------------------------------------------------------
  Copyright (C) 1994-2003  Hideaki Goto

        All Rights Reserved

  Permission to use, copy, modify, and distribute this software and
  its documentation for any purpose is hereby granted without fee,
  provided that (i) the above copyright notice and this permission
  notice appear in all copies and in supporting documentation, (ii)
  the name of the author, Hideaki Goto, may not be used in any
  advertising or otherwise to promote the sale, use or other
  dealings in this software without prior written authorization
  from the author, (iii) this software may not be used for
  commercial products without prior written permission from the
  author, and (iv) the notice of modification is specified in cases
  where modified copies of this software are distributed.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND.
  THE AUTHOR WILL NOT BE RESPONSIBLE FOR ANY DAMAGE CAUSED BY THIS
  SOFTWARE.
--------------------------------------------------------------------*/

#define	DESKEWING_ANGLE		0.001


//#define		_Mon1


#include	"SysConst.h"

#define		LRectPRs	8


#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<math.h>

#include	"utypes.h"
#include	"ufilep.h"
#include	"siplib.h"

#include	"TLobj.h"
#include	"Docfb.h"
#include	"TLimage.h"
#include	"TLfile.h"


extern "C" int	tlimage_filter(SIPImage *tlimage);




/*------------------------------------------------------
	Deskew image
------------------------------------------------------*/

ImgFBD *image_deskew(ImgFBD *image,double angle){
	TLImageH	TLImageH;
	ImgFBD		*tmpimage,*newimage;
	long		*rden;
	int		width,height;
	int		u,l;
	SIPRectangle	srect,drect;
	width  = (int)((double)image->width  * fabs(cos(angle)) \
	             + (double)image->height * fabs(sin(angle)));
	height = (int)((double)image->width  * fabs(sin(angle)) \
	             + (double)image->height * fabs(cos(angle)));
	if ( NULL == (rden = new long[height]) )  return(NULL);
	if ( NULL == (tmpimage = TLImageH.createfb(0,0,width,height,0,0)) ){
		delete []rden;
		return(NULL);
	}
	sip_rotate8(image->image,tmpimage->image, \
		tmpimage->width /2, tmpimage->height /2,angle,BOP_PUT);
	sip_projprofile(tmpimage->image,rden,0);
	for ( u=0 ; u<height ; u++ )  if ( rden[u] != 255 * width )  break;
	for ( l = height -1 ; l>=0 ; l-- )  if ( rden[l] != 255 * width )  break;
	if ( 0 >= (l = l - u + 1) ){  u = 0;  l = 1; }

	if ( NULL == (newimage = TLImageH.createfb(0,0,width,l,0,0)) ){
		delete []rden;
		TLImageH.deletefb(tmpimage);
		return(NULL);
	}
	srect.x = 0;  srect.y = u;  srect.width = width;  srect.height = l;
	drect.x = 0;  drect.y = 0;  drect.width = width;  drect.height = l;
	sip_CopyArea(tmpimage->image,&srect,newimage->image,&drect);
	TLImageH.deletefb(tmpimage);
	delete []rden;
	return(newimage);
}




/*------------------------------------------------------
	Image transposing
------------------------------------------------------*/

int image_ulrev(SIPImage *image){
	int	u,l;
	char	*lbuf;
	if ( 0 == (lbuf = new char [image->bytes_per_line]) )  return(-1);
	u = 0;  l = image->height -1;
	while ( u < l ){
		sip_GetScanLine(image,lbuf,0,u,image->width,image->depth);
		sip_PutScanLine(image,sip_getimgptr(image,l),0,u,image->width,image->depth);
		sip_PutScanLine(image,lbuf,0,l,image->width,image->depth);
		++u;  --l;
	}
	delete lbuf;
	return(0);
}




/*------------------------------------------------------
	String processing class
------------------------------------------------------*/


static int getimagesize(char *fname,int *width,int *height){
	int	retcode;
	d4pac	d4p;
	D4LOAD	D4LD;
	retcode = D4LD.fileopen(fname);
	if ( retcode != 0 )  return(retcode);
	D4LD.getupper(&d4p);
	if ( width  != 0 )  *width  = ((d4p.d4_data[0] + CWID) / CWID) * CWID;
	if ( height != 0 )  *height = ((d4p.d4_data[1] + CWID) / CWID) * CWID;
	D4LD.fileclose();
	return(0);
}


static void columnpack_all(TLGroup *TLG){
	int	i;
	TLOBJ	*TL;
	for ( i=0 ; i < TLG->lines ; i++ ){
		if ( 0 == (TL = TLG->tllist[i]) )  continue;
		tlproc_columnpack(TL);
	}
}


/*
static void print_lineinfo(STROBJ *Str){
	int		ix;
	lncoordinate	lncod;
	for ( ix=0 ; ix < Str->lines ; ix++ ){
		if ( 0 != Str->getcod(ix,&lncod) )  continue;
		printf("Line %d : (%d,%d)-(%d,%d) %d (%d)\n",ix, \
			lncod.x0,lncod.y0,lncod.x1,lncod.y1,Str->getboxes(ix), \
			Str->getvaboxes(ix));
	}
}
*/



/*------------------------------------------------------
	Message class
------------------------------------------------------*/

class MSG {
    public:
	void		usage(void);
	void		switcherr(void);
	void		readerr(char *name);
	void		writeerr(char *name);
	void		nonsupported(char *name);
	void		memoryover(void);
	void		sizemismatch(void);
};
    	
    	
void MSG :: usage(void){		// Display Usage
	puts("strextr  v2.4, Copyright (C) 1994-2003 Hideaki Goto ");
	puts("usage : strextr [-options] str_file [-image image_file]");
	puts("                [-resdir result_dir] [-o out_str_file]");
	puts("          -dpi N     : set image resolution as N=200,400 (defult: 200dpi)");
	puts("          -trans     : image transposing ");
	puts("          -raw       : raw extraction mode ");
	puts("          -rect      : rectangular extraction mode ");
	puts("          -irect     : inclined-rectangular extraction mode ");
	puts("        * -lrect     : local-rectangular extraction mode ");
	puts("          -intrp     : PR-interpolation ON ");
	puts("          -filter    : apply image filter ");
	puts("          -deskew    : deskew text lines ");
	puts("          -nwr       : suppress writing file ");
	puts("          -verbose   : display process information ");
//	puts("          -quiet     : no Conversion-Info. display ");
}


void MSG :: switcherr(void){
	puts("Switch error.");
}


void MSG :: readerr(char *name){
	printf("Can't read %s.\n",name);
}


void MSG :: writeerr(char *name){
	printf("Can't write %s.\n",name);
}


void MSG :: nonsupported(char *name){
	printf("Non-supported file %s.\n",name);
}


void MSG :: memoryover(void){
	puts("Not enough memory.");
}


void MSG :: sizemismatch(void){
	puts("Size mismatch.");
}




/*------------------------------------------------------
	Argument / Option class
------------------------------------------------------*/

class ARGS {
    private:
	char		*fname;
	char		*out_fname;
	char		resfname[128];
	char		*fnameimg;
	char		*dirres;
    	int		sw_verbose;
	int		sw_rect;
	int		sw_intrp;
	int		sw_dpi;
	int		sw_nowrite;
	int		sw_trans;
	int		sw_filter;
	int		sw_deskew;
	int		rfcount;
    public:
	int		scanargv(int ac,char **av);
	int		SWverbose(void) {  return sw_verbose; }
	int		SWrect(void) {  return sw_rect; }
	int		SWintrp(void) {  return sw_intrp; }
	int		SWdpi(void) {  return sw_dpi; }
	int		SWnowrite(void) {  return sw_nowrite; }
	int		SWtrans(void) {  return sw_trans; }
	int		SWfilter(void) {  return sw_filter; }
	int		SWdeskew(void) {  return sw_deskew; }
	char		*imgfname(void) {  return fnameimg; }
	char		*resdir(void) {  return dirres; }
	char		*infile(void) {  return fname; }
	char		*outfile(void) {  return out_fname; }
	char		*resfilecur(void){  return resfname; }
	char		*resfilenext(void);
			ARGS(void);
};


ARGS :: ARGS(void){
	sw_verbose = 0;
	sw_rect = 3;
	sw_intrp = 0;
	sw_dpi = 200;
	sw_nowrite = 0;
	sw_trans = 0;
	sw_filter = 0;
	sw_deskew = 0;
	fname = out_fname = 0;
	fnameimg = "tmp.img";
	dirres = "./cuttmp";
	rfcount = 0;
}


int ARGS :: scanargv(int ac,char **av){
	int	i1,err,skip;
	char	*sw;
	err = 0;
	skip = 0;
	--ac;
	for( i1 = 0 ; i1 < ac ; ++i1 ){
		if( skip ){  skip = 0;  continue;  }
		if( *( sw = av[i1+1] ) == '-' ){
			++sw;
			if( 0 == strcmp(sw,"o") ){
				if((i1+1)<ac){
					out_fname = av[i1+2];
					skip = -1;
					continue;
				}
				else{	err = -1;  break;  }
			}
			if( 0 == strcmp(sw,"image") ){
				if((i1+1)<ac){
					fnameimg = av[i1+2];
					skip = -1;
					continue;
				}
				else{	err = -1;  break;  }
			}
			if( 0 == strcmp(sw,"resdir") ){
				if((i1+1)<ac){
					dirres = av[i1+2];
					skip = -1;
					continue;
				}
				else{	err = -1;  break;  }
			}
			if( 0 == strcmp(sw,"dpi") ){
				if((i1+1)<ac){
					sw_dpi = atoi(av[i1+2]);
					if ( sw_dpi != 400 && sw_dpi != 200 ){
						err = -1;  break; 
					}
					skip = -1;
					continue;
				}
				else{	err = -1;  break;  }
			}
			if( 0 == strcmp(sw,"raw") ){	sw_rect = 0;  continue; }
			if( 0 == strcmp(sw,"rect") ){	sw_rect = 1;  continue; }
			if( 0 == strcmp(sw,"irect") ){	sw_rect = 2;  continue; }
			if( 0 == strcmp(sw,"lrect") ){	sw_rect = 3;  continue; }
			if( 0 == strcmp(sw,"intrp") ){	sw_intrp = -1;  continue; }
			if( 0 == strcmp(sw,"nwr") ){	sw_nowrite = -1;  continue; }
			if( 0 == strcmp(sw,"trans") ){	sw_trans = -1;  continue; }
			if( 0 == strcmp(sw,"filter") ){	sw_filter = -1;  continue; }
			if( 0 == strcmp(sw,"deskew") ){	sw_deskew = -1;  continue; }
			if( 0 == strcmp(sw,"quiet") ){	sw_verbose = 0;  continue; }
			if( 0 == strcmp(sw,"verbose") ){  sw_verbose = -1;  continue; }
			if( 0 == strcmp(sw,"v") ){  sw_verbose = -1;  continue; }
			err = -1;  break;
		}
		else{	fname = av[i1+1];
		}
	}
	return(err);
}


char * ARGS :: resfilenext(void){
	char	ns[10];
	strncpy(resfname,dirres,120);
	sprintf(ns,"/STR%04d",(rfcount++));
	strcat(resfname,ns);
	return(resfname);
}




/*------------------------------------------------------
	Main routine
------------------------------------------------------*/

int main(int ac,char **av){
	ARGS		ARGS;
	MSG		MSG;
	TLGroup		TLG;
	TLFILE		TLFILE;
	TLImageH	TLImageH;
	DOCFB		DOCFB;
	int		width,height;
	int		retcode,errflag;
	int		ix;
	ImgFBD		docimg;
	ImgFBD		*strimg;
	int		multiply;
	int		out_serial;

	if ( ac < 2 ){  MSG.usage();  exit(0);  }

	if ( 0 != ARGS.scanargv(ac,av) ){
		MSG.switcherr();
		exit(1);
	}

	if ( ARGS.SWdpi() != 400 )  multiply = 1;  else  multiply = 2;


//-----------	Load String Data and Pack Column-Data

	if ( 0 != TLFILE.load(ARGS.infile(),&TLG) ){
		MSG.readerr(ARGS.infile());  exit(2);
	}
	columnpack_all(&TLG);
	if ( ARGS.SWverbose() ){
		printf ("%d lines:\n",TLG.lines);
#ifdef _Mon1 
		TLG.print_lineinfo();
#endif
	}



//-----------	Load Image File

	DOCFB.setpadding(BANDWIDTH * multiply,BANDWIDTH * multiply);
	DOCFB.setbgcolor(0xff);
	if ( ! ARGS.SWtrans() )
		retcode = DOCFB.load_file(ARGS.imgfname(),0,0);
	else	retcode = DOCFB.load_file(ARGS.imgfname(),0,1);
	switch ( retcode ){
	  case 0:	break;
	  case -1:	MSG.memoryover();  exit(3);
	  case -3:	MSG.nonsupported(ARGS.imgfname());  exit(2);
	  default:	MSG.readerr(ARGS.imgfname());  exit(2);
	}

	docimg.xoff = docimg.yoff = 0;
	docimg.width = DOCFB.width;
	docimg.height = DOCFB.height;
	docimg.image = DOCFB.image;


//-----------	Get Image Size

	width  = DOCFB.width;
	height = DOCFB.height;

	if ( ARGS.SWverbose() ){
		printf("Image  size : %d x %d\n",width,height);
	}
	if ( ARGS.SWtrans() ){
		int	stmp;
		stmp = width;  width = height;  height = stmp;
	}


//-----------	String image extraction

	errflag = 0;
	for ( ix = out_serial = 0 ; ix < TLG.lines ; ix++ ){
		if ( ARGS.SWverbose() )  printf("Line %d (%d) : ",ix,out_serial);
		if ( (TLG.tllist[ix]->stat & TLSTAT_ACTIVE) == 0 )  continue;

		if ( ARGS.SWintrp() || (ARGS.SWrect() >= 2) ){
			TLOBJ	*TL = TLG.tllist[ix];
			if ( TL == 0 )  continue;
			tlproc_interp(TL,CWID);
		}
		strimg = TLImageH.getimage(&docimg,&TLG,ix,CWID,multiply, \
					LRectPRs,ARGS.SWrect());
		if ( strimg == 0 ){
			MSG.memoryover();
			exit(3);
		}

		if ( ARGS.SWfilter() ){
			int	f;
			f = tlimage_filter(strimg->image);
			if ( f < 0 ){ MSG.memoryover(); exit(3); }
			if ( f == 0 ){
				TLG.removeline(ix);
				if ( ARGS.SWverbose() )  printf("rejected.\n");
				TLImageH.deletefb(strimg);
				continue;
			}
		}

		if ( ARGS.SWdeskew() ){
			ImgFBD	*tmpimg;
			double	angle;
			angle = TLImageH.getskew(&TLG,ix);
			if ( fabs(angle) > DESKEWING_ANGLE ){
				tmpimg = image_deskew(strimg, angle);
				if ( NULL == tmpimg ){
					MSG.memoryover();
					exit(3);
				}
				TLImageH.deletefb(strimg);
				strimg = tmpimg;
			}
		}

		if ( ! ARGS.SWnowrite() ){
			if ( ARGS.SWtrans() ){
				image_ulrev(strimg->image);
			}
			if ( 0 != TLImageH.save(ARGS.resfilenext(),strimg) ){
				TLImageH.deletefb(strimg);
				MSG.writeerr(ARGS.resfilecur());
				exit(5);
			}
		}
		TLImageH.deletefb(strimg);
		if ( ARGS.SWverbose() )  printf("done.\n");
		++out_serial;
	}


//-----------	Save String Data

	if ( ARGS.outfile() ){
		if ( 0 != TLFILE.save(ARGS.outfile(),&TLG,0) ){
			MSG.writeerr(ARGS.outfile());  exit(2);
		}
	}

	return(0);
}
