/*--------------------------------------------------------------
	Finds orientations of local text regions   txtori v1.32
		Written  by H.Goto,  May 1995
		Modified by H.Goto,  Aug 1995
		Modified by H.Goto,  Nov 1997
		Rewritten by H.Goto, Jul 1998 (for project-O2)
		Modified by H.Goto,  Jul 2003
--------------------------------------------------------------*/

/*--------------------------------------------------------------------
  Copyright (C) 1995-2003  Hideaki Goto

        All Rights Reserved

  Permission to use, copy, modify, and distribute this software and
  its documentation for any purpose is hereby granted without fee,
  provided that (i) the above copyright notice and this permission
  notice appear in all copies and in supporting documentation, (ii)
  the name of the author, Hideaki Goto, may not be used in any
  advertising or otherwise to promote the sale, use or other
  dealings in this software without prior written authorization
  from the author, (iii) this software may not be used for
  commercial products without prior written permission from the
  author, and (iv) the notice of modification is specified in cases
  where modified copies of this software are distributed.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND.
  THE AUTHOR WILL NOT BE RESPONSIBLE FOR ANY DAMAGE CAUSED BY THIS
  SOFTWARE.
--------------------------------------------------------------------*/


#define		CWID		16		/* Column width */
#define		CellSize	(CWID * 4)	/* = 64 */
#define		LocalRects	50

#define		RectHeightThin	5	/* = 1.81pt (at 200dpi) */


#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<math.h>
#include	<unistd.h>

#include	"ufilep.h"
#include	"imgobj.h"
#include	"sgplib.h"
#include	"stdps.h"




/*------------------------------------------------------
	Global Constants
------------------------------------------------------*/

#define		LRGSTAT_Null	0
#define		LRGSTAT_Hrz	0x01
#define		LRGSTAT_Vrt	0x02
#define		LRGSTAT_Both	(LRGSTAT_Hrz | LRGSTAT_Vrt)
#define		LRGSTAT_Unknown	0x80

#define		RECTSTAT_Thin	0x01




/*------------------------------------------------------
	Local Rectangle Group
------------------------------------------------------*/

class LRGrp {
    private:
	int		maxoris;
	CRects		*Rects;

	int		*actcounts;
	void		clearcache(int ori);

    public:
	int		status;
	int		R;
	double		angle;
	double		LH,LW;
	int		addrect(int ori,int x0,int y0,int w,int h);
	CRects		*getrects(int ori);
	int		count(int ori);
	int		totalcount(void);
	int		actcount(int ori);
	int		setattr(int ori,int crn,int attr);
	int		getattr(int ori,int crn);
	int		setattrall(int ori,int attr);
	int		init(int oris);
			LRGrp();
	virtual		~LRGrp();
};


LRGrp :: LRGrp(){
	maxoris = 0;
	Rects = 0;
	actcounts = 0;
}


LRGrp :: ~LRGrp(){
	if ( Rects != 0 )  delete []Rects;
	if ( actcounts != 0 )  delete []actcounts;
}


int LRGrp :: init(int oris){
	int	i;
	if ( Rects != 0 )  delete []Rects;
	if ( actcounts != 0 )  delete []actcounts;
	if ( 0 == (Rects = new CRects[oris]) )  return(-1);
	if ( 0 == (actcounts = new int[oris]) )  return(-1);
	for ( i=0 ; i<oris ; i++ )  actcounts[i] = -1;
	maxoris = oris;
	status = LRGSTAT_Unknown;
	return(0);
}


void LRGrp :: clearcache(int ori){
	actcounts[ori] = -1;
	return;
}


int LRGrp :: addrect(int ori,int x0,int y0,int w,int h){
	if ( ori >= maxoris )  return(-1);
	clearcache(ori);
	return( Rects[ori].create(x0,y0,x0 + w -1,y0 + h -1) );
}


CRects * LRGrp :: getrects(int ori){
	if ( ori >= maxoris )  return(0);
	return( &Rects[ori] );
}


int LRGrp :: count(int ori){
	if ( ori >= maxoris )  return(-1);
	return( Rects[ori].counts );
}


int LRGrp :: totalcount(void){
	int	i,tc;
	tc = 0;
	for ( i=0 ; i<maxoris ; i++ )  tc += Rects[i].counts;
	return(tc);
}


int LRGrp :: actcount(int ori){
	int	crn,acts;
	int	*attr;
	if ( ori >= maxoris )  return(-1);
	if ( actcounts[ori] != -1 )  return(actcounts[ori]);
	for ( crn=acts=0 ; crn<Rects[ori].counts ; crn++ ){
		if ( ! Rects[ori].isactive(crn) )  continue;
		attr = (int *)Rects[ori].getattrp(crn);
		if ( attr != 0 ){
			if ( *attr != 0 )  continue;
		}
		++acts;
	}
	actcounts[ori] = acts;
	return( acts );
}


int LRGrp :: setattr(int ori,int crn,int attr){
	int	*attr0;
	if ( ori >= maxoris )  return(-1);
	if ( crn >= Rects[ori].counts )  return(-1);
	clearcache(ori);
	attr0 = (int *)Rects[ori].getattrp(crn);
	if ( 0 == attr0 ){
		if ( 0 == (attr0 = new int) )  return(-1);
		Rects[ori].setattrp(crn,(void *)attr0);
	}
	*attr0 = attr;
	return(0);
}


int LRGrp :: getattr(int ori,int crn){
	int	*attr;
	if ( ori >= maxoris )  return(-1);
	if ( crn >= Rects[ori].counts )  return(-1);
	attr = (int *)Rects[ori].getattrp(crn);
	if ( 0 == attr )  return(-1);
	return(*attr);
}


int LRGrp :: setattrall(int ori,int attr){
	int	crn;
	if ( ori >= maxoris )  return(-1);
	for ( crn = 0 ; crn < Rects[ori].counts ; crn++ ){
		setattr(ori,crn,attr);
	}
	return(0);
}




/*------------------------------------------------------
	Local Rectangle Group Matrix
------------------------------------------------------*/

#define		HistMax		4.0
#define		HistDiv		0.2
#define		HistLen		((int)((HistMax / HistDiv) +1))
#define		ClipSpan	(log(3.0/2.0))
#define		AHistDiv	2.0
#define		AHistLen	((int)(90.0 / AHistDiv) +2)
#define		AHistOff	(AHistLen /2)
#define		DistMul		2.0
#define		DistMulCR	2.4

#define		Tangle		(16 /(AHistDiv * AHistDiv))
#define		AWinHalf	(30 / (int)AHistDiv)
#define		AWinSize	(AWinHalf * 2) + 1)


class LRGMatrix{
    private:
	int		*lhist;
	int		*ahist,*ahist90,*ahistsum;
	LRGrp		*LRG;
	CRects		*LocRects;
	int		setlocalrects(int mx,int my);
	int		makehist(int ori);
	void		clearanglehist();
	int		makeanglehist(int rectmode,int ori,double distlimit);
	int		peakpos(peaklisti *peak);
	double		stdrectsize(int ori);
	int		pickattnrects(int ori,double lhlow,double lhhi);
	double		meanangle();
    public:
	int		mwidth,mheight,msize;
	int		oris;
	int		alloc(int mwidth,int mheight,int oris);
	int		initallcells();

	int		getrects(int ori,int mx,int my,int r);
	int		getactrects(int ori,int mx,int my,int r);
	LRGrp		*cell(int mx,int my){  return( &LRG[mx + (my * mwidth)] );  }
	int		setattrall(int attr);

	int		setradius(int rects);
	int		findorientation(int mx,int my,int rectmode,int verbose);

			LRGMatrix();
	virtual		~LRGMatrix();
};


LRGMatrix :: LRGMatrix(){
	lhist = ahist = 0;
	LRG = 0;
	LocRects = 0;
	mwidth = mheight = msize = 0;
}


LRGMatrix :: ~LRGMatrix(){
	if ( lhist != 0 )  delete []lhist;
	if ( ahist != 0 )  delete []ahist;
	if ( LRG != 0 )  delete []LRG;
	if ( LocRects != 0 )  delete []LocRects;
}


int LRGMatrix :: setlocalrects(int mx,int my){
	int	ori,x,y,crn,r;
	int	xs,ys,xe,ye;
	CRects	*Rects;
	CRect	*rect;
	int	*attr;

	for ( ori=0 ; ori<oris ; ori++ )  LocRects[ori].clear();
	r = cell(mx,my)->R;
	if ( r < 0 )  return(0);

	if ( 0 > (xs = mx - r) )  xs = 0;
	if ( 0 > (ys = my - r) )  ys = 0;
	if ( mwidth  <= (xe = mx + r) )  xe = mwidth  -1;
	if ( mheight <= (ye = my + r) )  ye = mheight -1;
	for ( ori=0 ; ori<oris ; ori++ ){
		for ( y=ys ; y<=ye ; y++ ){
			for ( x=xs ; x<=xe ; x++ ){
				Rects = cell(x,y)->getrects(ori);
				for ( crn=0 ; crn<Rects->counts ; crn++ ){
					if ( ! Rects->isactive(crn) )  continue;
					if ( 0 != (attr = (int *)Rects->getattrp(crn)) ){
						if ( *attr != 0 )  continue;
					}
					rect = Rects->getrect(crn);
					if ( 0 > LocRects[ori].create(rect->x1,rect->y1,rect->x2,rect->y2) )  return(-1);
				}
			}
		}
	}
	return(0);
}


int LRGMatrix :: makehist(int ori){
	int	i;
	double	p;
	for ( i=0 ; i < HistLen ; i++ )  lhist[i] = 0;
	for ( i=0 ; i < LocRects[ori].counts ; i++ ){
		p = log((double)LocRects[ori].getrect(i)->height());
		if ( (p < 0) || (p >= HistMax) )  continue;
		++lhist[(int)(p / HistDiv)];
	}
	return(0);
}


void LRGMatrix :: clearanglehist(){
	int	i;
	for ( i=0 ; i < AHistLen ; i++ )  ahist[i] = ahist90[i] = 0;
}


int LRGMatrix :: makeanglehist(int rectmode,int ori,double distlimit){
	int	i,j;
	int	cx1,cy1,cx2,cy2;
	int	dx,dy;
	CRect	*rect;
	CRects	*LRs;
	double	a;
	double	deg;
	int	dl2;

	dl2 = (int)(distlimit * distlimit);
	deg = 45 / atan(1.0);
	LRs = &LocRects[ori];
	for ( i=0 ; i < LocRects[ori].counts ; i++ ){
		rect = LocRects[ori].getrect(i);
		if ( ori == 0 ){
			cx1 = (rect->x1 + rect->x2) /2;
			cy1 = (rect->y1 + rect->y2) /2;
		}
		else{	cy1 = (rect->x1 + rect->x2) /2;
			cx1 = (rect->y1 + rect->y2) /2;
		}
		for ( j=i+1 ; j < LRs->counts ; j++ ){
			rect = LRs->getrect(j);
			if ( ori == 0 ){
				cx2 = (rect->x1 + rect->x2) /2;
				cy2 = (rect->y1 + rect->y2) /2;
			}
			else{	cy2 = (rect->x1 + rect->x2) /2;
				cx2 = (rect->y1 + rect->y2) /2;
			}
			dx = cx2 - cx1;
			dy = cy2 - cy1;
			if ( dx < -(int)distlimit || (int)distlimit < dx )  continue;
			if ( dy < -(int)distlimit || (int)distlimit < dy )  continue;
			if ( (dx * dx + dy * dy) > dl2 )  continue;
			if ( abs(dy) <= abs(dx) ){
				if ( (rectmode == 0) && (ori == 1) )  continue;
				if ( dx == 0 )  continue;
				a = deg * atan((double)dy / (double)dx);
//				if ( a < -AWinHalf || a > AWinHalf )  continue;
				++ahist[ (int)(a / AHistDiv) + AHistOff ];
			}
			else{	if ( (rectmode == 0) && (ori == 0) )  continue;
				if ( dy == 0 )  continue;
				a = -deg * atan((double)dx / (double)dy);
//				if ( a < -AWinHalf || a > AWinHalf )  continue;
				++ahist90[ (int)(a / AHistDiv) + AHistOff ];
			}
		}
	}
	return(0);
}


int LRGMatrix :: peakpos(peaklisti *peak){
	int		peaks,pos,pos2;
	peaklisti	*pks;
	peaklisti	pktmp;
	if ( 0 == (pks = new peaklisti[HistLen]) )  return(-1);
	peaks = sgp_getpeaksi(lhist,pks,(int)HistLen,(int)HistLen,0,32767);
	if ( peaks > 0 ){
		for ( pos=0 ; pos<peaks ; pos++ ){
			for ( pos2=pos+1 ; pos2<peaks ; pos2++ ){
				if ( pks[pos].value < pks[pos2].value ){
					pktmp = pks[pos];
					pks[pos] = pks[pos2];
					pks[pos2] = pktmp;
				}
			}
		}
		if ( peaks > 1 ){
			if ( (pks[1].value *2) > pks[0].value )  pks[0] = pks[1];
		}
	}
	else{	pks[0].pos   = 0;
		pks[0].value = 0;
	}
	if ( 0 != peak )  *peak = pks[0];
	delete []pks;
	return(pks[0].pos);
}


double LRGMatrix :: stdrectsize(int ori){
	int	pos;
	makehist(ori);
	if ( 0 >= (pos = peakpos(0)) )  return(0);
	return(((double)pos + 0.5) * HistDiv);
}


int LRGMatrix :: pickattnrects(int ori,double lhlow,double lhhi){
	int	i;
	int	h;
	double	lh;
	for ( i=0 ; i < LocRects[ori].counts ; i++ ){
		h = LocRects[ori].getrect(i)->height();
		lh = log((double)h);
		if ( (lh < lhlow) || (lhhi < lh) )  LocRects[ori].destroy(i);
	}
	return( LocRects[ori].truncate() );
}


double LRGMatrix :: meanangle(){
	int		i,peaks;
	int		maxpkh;
	int		mask[] = {1,2,1};
	peaklisti	*pks;
	int		meanpos;
	for ( i=0 ; i<AHistLen ; i++ ){
		ahistsum[i] = ahist[i] + ahist90[i];
	}
	sgp_convolutei(ahistsum,ahistsum,AHistLen,0,mask,1,3);
	if ( 0 == (pks = new peaklisti[AHistLen]) )  return(-1);
	peaks = sgp_getpeaksi(ahistsum,pks,(int)AHistLen,(int)AHistLen,0,32767);
	meanpos = AHistOff;
	maxpkh = 0;
	for ( i=0 ; i<peaks ; i++ ){
		if ( pks[i].value > maxpkh ){
			meanpos = pks[i].pos;
			maxpkh = pks[i].value;
		}
	}
	delete []pks;
	return((double)meanpos);
}


int LRGMatrix :: alloc(int mwidth,int mheight,int oris){
	int	size;
	size = mwidth * mheight;
	if ( lhist == 0 ){
		if ( 0 == (lhist = new int[HistLen]) )  return(-1);
	}
	if ( ahist == 0 ){
		if ( 0 == (ahist = new int[AHistLen *3]) )  return(-1);
		ahist90 = ahist + AHistLen;
		ahistsum = ahist90 + AHistLen;
	}
	if ( LRG != 0 ){
		delete []LRG;
		mwidth = mheight = msize = 0;
	}
	if ( LocRects != 0 ){
		delete []LocRects;  LocRects = 0;
	}
	if ( 0 == (LRG = new LRGrp[size]) )  return(-1);
	if ( 0 == (LocRects = new CRects[oris]) )  return(-1);
	LRGMatrix :: mwidth  = mwidth;
	LRGMatrix :: mheight = mheight;
	LRGMatrix :: msize   = size;
	LRGMatrix :: oris    = oris;
	return(0);
}


int LRGMatrix :: initallcells(){
	int	i;
	for ( i=0 ; i<msize ; i++ ){
		if ( -1 == LRG[i].init(oris) )  return(-1);
	}
	return(0);
}


int LRGMatrix :: getrects(int ori,int mx,int my,int r){
	int	x,y;
	int	xs,ys,xe,ye;
	int	rects;
	if ( 0 > (xs = mx - r) )  xs = 0;
	if ( 0 > (ys = my - r) )  ys = 0;
	if ( mwidth  <= (xe = mx + r) )  xe = mwidth  -1;
	if ( mheight <= (ye = my + r) )  ye = mheight -1;
	rects = 0;
	for ( y=ys ; y<=ye ; y++ ){
		for ( x=xs ; x<=xe ; x++ ){
			rects += cell(x,y)->count(ori);
		}
	}
	return(rects);
}


int LRGMatrix :: getactrects(int ori,int mx,int my,int r){
	int	x,y;
	int	xs,ys,xe,ye;
	int	rects;
	if ( 0 > (xs = mx - r) )  xs = 0;
	if ( 0 > (ys = my - r) )  ys = 0;
	if ( mwidth  <= (xe = mx + r) )  xe = mwidth  -1;
	if ( mheight <= (ye = my + r) )  ye = mheight -1;
	rects = 0;
	for ( y=ys ; y<=ye ; y++ ){
		for ( x=xs ; x<=xe ; x++ ){
			rects += cell(x,y)->actcount(ori);
		}
	}
	return(rects);
}


int LRGMatrix :: setattrall(int attr){
	int	i,ori;
	for ( i = 0 ; i < msize ; i++ ){
		for ( ori = 0 ; ori < oris ; ori++ ){
			if ( LRG[i].setattrall(ori,attr) )  return(-1);
		}
	}
	return(0);
}


int LRGMatrix :: setradius(int rects){
	int	lf;
	int	ori;
	int	mx,my,r;
	for ( my = 0 ; my < mheight ; my++ ){
		for ( mx = 0 ; mx < mwidth ; mx++ ){
			cell(mx,my)->R = -1;
			lf = -1;
			/* Note: maximum window size = 5x5 */
			for ( r = 0 ; (r < 3) && lf ; r++ ){
				for ( ori=0 ; ori<oris ; ori++ ){
					if ( getactrects(ori,mx,my,r) >= rects ){
						lf = 0;
						cell(mx,my)->R = r;
						break;
					}
				}
			}
		}
	}
	return(0);
}


int LRGMatrix :: findorientation(int mx,int my,int rectmode,int verbose){
	int	rectsH,rectsV;
	double	lsizeH,lsizeV;
	double	hmeanH,hmeanV,hmeanS;
	double	sdH,sdV;
	int	ahs,ahe;

	if ( cell(mx,my)->R < 0 ){	/* rectangles too few */
		return(LRGSTAT_Both);
	}

	if ( setlocalrects(mx,my) )  return(-1);
	clearanglehist();

	lsizeH = stdrectsize(0);
	rectsH = pickattnrects(0,lsizeH - ClipSpan,lsizeH + ClipSpan);
	lsizeV = stdrectsize(1);
	rectsV = pickattnrects(1,lsizeV - ClipSpan,lsizeV + ClipSpan);
	if ( 0 == rectmode ){
		makeanglehist(0,0,CWID * DistMul);		/* PR */
		makeanglehist(0,1,CWID * DistMul);
	}
	else{	makeanglehist(1,0,exp(lsizeH) * DistMulCR);	/* CR */
		makeanglehist(1,1,exp(lsizeV) * DistMulCR);
	}

	hmeanS = meanangle();
	cell(mx,my)->angle = (hmeanS - AHistOff) * AHistDiv;
//	printf("S-Angle %f deg\n",(hmeanS - AHistOff) * AHistDiv);

	hmeanH = sgp_gethistmeani(ahist,AHistLen);
	hmeanH = hmeanS;
	if ( 0 > (ahs = (int)(0.5 + hmeanH) - AWinHalf) )  ahs = 0;
	if ( AHistLen <= (ahe = (int)(0.5 + hmeanH) + AWinHalf) )  ahe = AHistLen -1;
	hmeanH = sgp_gethistmeani(ahist + ahs,ahe - ahs +1) + (double)ahs;
	sdH = sgp_gethistsdi(ahist + ahs,ahe - ahs +1,(hmeanH - ahs));
//	cell(mx,my)->angle = (hmeanH - AHistOff) * AHistDiv;
//	printf("H-Angle %f deg\n",cell(mx,my)->angle);

	hmeanV = sgp_gethistmeani(ahist90,AHistLen);
	hmeanV = hmeanS;
	if ( 0 > (ahs = (int)(0.5 + hmeanV) - AWinHalf) )  ahs = 0;
	if ( AHistLen <= (ahe = (int)(0.5 + hmeanV) + AWinHalf) )  ahe = AHistLen -1;
	hmeanV = sgp_gethistmeani(ahist90 + ahs,ahe - ahs +1) + (double)ahs;
	sdV = sgp_gethistsdi(ahist90 + ahs,ahe - ahs +1,(hmeanV - ahs));

	if ( verbose ){
		printf("H: %f  ( %d-> %d)\n",lsizeH,LocRects[0].counts,rectsH);
		printf("%f  %f\n",(hmeanH - AHistOff) * AHistDiv,sdH);
		printf("V: %f  ( %d-> %d)\n",lsizeV,LocRects[1].counts,rectsV);
		printf("%f  %f\n",(hmeanV - AHistOff) * AHistDiv,sdV);
	}

	if ( sdH < Tangle && sdV >= Tangle )  return(LRGSTAT_Hrz);
	if ( sdV < Tangle && sdH >= Tangle )  return(LRGSTAT_Vrt);

	return(LRGSTAT_Both);
}




/*------------------------------------------------------
	postscript file export class
------------------------------------------------------*/

class PSEXP {
    private:
	FILE		*fp;
    public:
	int		open(FILE *fp);
	int		line(int x0,int y0,int x1,int y1,int width);
	int		circle(int x0,int y0,int r,int width);
	int		orimark(int markheight,int x,int y,int ori,double angle);
	int		close(void);
			PSEXP(void);
};


PSEXP :: PSEXP(void){
	fp = stdout;
}


int PSEXP :: open(FILE *fp){
	PSEXP::fp = fp;
	stdps_setsize("A4");
	return( stdps_header(fp) );
}


int PSEXP :: line(int x0,int y0,int x1,int y1,int width){
	fprintf(fp,"%d setlinewidth\n",width);
	fprintf(fp,"newpath %d %d moveto %d %d lineto  stroke\n",x0,y0,x1,y1);
	return(0);
}


int PSEXP :: circle(int x0,int y0,int r,int width){
	fprintf(fp,"%d setlinewidth\n",width);
	fprintf(fp,"newpath %d %d %d 0 360 arc stroke\n",x0,y0,r);
	return(0);
}


int PSEXP :: orimark(int markheight,int x,int y,int ori,double angle){
	int	xoff,yoff;
	xoff = (int)((double)(markheight /2) * cos(angle));
	yoff = (int)((double)(markheight /2) * sin(angle));
	if ( ori & 0x01 )  line(x-xoff,y-yoff,x+xoff,y+yoff,3);
	if ( ori & 0x02 )  line(x+yoff,y-xoff,x-yoff,y+xoff,3);
	return(0);
}


int PSEXP :: close(void){
	return( stdps_end(fp) );
}




/*------------------------------------------------------
	Export direction matrix to Postscript file
------------------------------------------------------*/

int export_direction(char *fname,LRGMatrix *LRGMtx){
	int	mx,my,xyoff;
	int	cellstat;
	PSEXP	PSEXP;
	FILE	*fp;
	double	rad;
	rad = atan(1) / 45;
	xyoff = CellSize /2;
	if ( 0 == (fp = fopen(fname,"w")) )  return(-1);
	if ( 0 != PSEXP.open(fp) )  return(-1);
	for ( my = 0 ; my < LRGMtx->mheight ; my++ ){
		for ( mx = 0 ; mx < LRGMtx->mwidth ; mx++ ){
			cellstat = LRGMtx->cell(mx,my)->status;
			if ( 0 == LRGMtx->cell(mx,my)->totalcount() )  cellstat = LRGSTAT_Null;
			switch ( cellstat ){
			  case LRGSTAT_Null:
				break;
			  case LRGSTAT_Hrz:
				PSEXP.orimark(CellSize *2 /3,mx * CellSize + xyoff,my * CellSize + xyoff,0x01,rad * LRGMtx->cell(mx,my)->angle);
				break;
			  case LRGSTAT_Vrt:
				PSEXP.orimark(CellSize *2 /3,mx * CellSize + xyoff,my * CellSize + xyoff,0x02,rad * LRGMtx->cell(mx,my)->angle);
				break;
			  case LRGSTAT_Both:
				PSEXP.orimark(CellSize *2 /3,mx * CellSize + xyoff,my * CellSize + xyoff,0x01,rad * LRGMtx->cell(mx,my)->angle);
				PSEXP.orimark(CellSize *2 /3,mx * CellSize + xyoff,my * CellSize + xyoff,0x02,rad * LRGMtx->cell(mx,my)->angle);
//				PSEXP.circle(mx * CellSize + xyoff,my * CellSize + xyoff,CellSize *1 /3,3);
				break;
			  default:
				PSEXP.circle(mx * CellSize + xyoff,my * CellSize + xyoff,CellSize *1 /6,3);
				break;
			}
		}
	}
	if ( 0 != PSEXP.close() )  return(-1);
	return(0);
}




/*------------------------------------------------------
	Print direction matrix
------------------------------------------------------*/

int print_direction(LRGMatrix *LRGMtx){
	int	mx,my;
	int	cellstat;
	for ( my = 0 ; my < LRGMtx->mheight ; my++ ){
		for ( mx = 0 ; mx < LRGMtx->mwidth ; mx++ ){
			cellstat = LRGMtx->cell(mx,my)->status;
			switch ( cellstat ){
			  case LRGSTAT_Null:
				fputs(" .",stdout);
				break;
			  case LRGSTAT_Hrz:
				fputs(" -",stdout);
				break;
			  case LRGSTAT_Vrt:
				fputs(" |",stdout);
				break;
			  case LRGSTAT_Both:
				fputs(" +",stdout);
				break;
			  default:
				fputs(" ?",stdout);
				break;
			}
		}
		fputs("\n",stdout);
	}
	return(0);
}




/*------------------------------------------------------
	Noise reduction
------------------------------------------------------*/

int NR_area(LRGMatrix *LRGMtx){
	int	mx,my,mxs,mys;
	int	cellstat,*cstmp;
	int	i,j,r,sv;
	int	mwidth,mheight;
	mwidth  = LRGMtx->mwidth;
	mheight = LRGMtx->mheight;
	if ( 0 == (cstmp = new int[mwidth * mheight]) )  return(-1);

				/* Copy status matrix and find null regions */
	for ( my=0 ; my<mheight ; my++ ){
		for ( mx=0 ; mx<mwidth ; mx++ ){
			cellstat = LRGMtx->cell(mx,my)->status;
			if ( LRGMtx->cell(mx,my)->totalcount() == 0 )  cellstat = LRGSTAT_Null;
			cstmp[mx + (my * mwidth)] = cellstat;
		}
	}

	for ( my = 0 ; my < mheight ; my++ ){
		for ( mx = 0 ; mx < mwidth ; mx++ ){
			cellstat = cstmp[mx + (my * mwidth)];
			if ( cellstat == LRGSTAT_Null )  continue;
			r = LRGMtx->cell(mx,my)->R;
			sv = LRGSTAT_Null;
			for ( j = -r ; j <= r ; j++ ){
				mys = my + j;
				if ( mys <  0       )  continue;
				if ( mys >= mheight )  continue;
				mys *= mwidth;
				if ( sv == LRGSTAT_Both )  break;
				for ( i = -r ; i <= r ; i++ ){
					if ( (i == 0) && (j == 0) )  continue;
					mxs = mx + i;
					if ( mxs <  0      )  continue;
					if ( mxs >= mwidth )  continue;
					if ( cstmp[mxs + mys] == LRGSTAT_Null )  continue;
					if ( sv == LRGSTAT_Null ){
						sv = cstmp[mxs + mys];
						if ( sv == LRGSTAT_Both )  break;
						continue;
					}
					if ( sv != cstmp[mxs + mys] ){
						sv = LRGSTAT_Both;  break;
					}
				}
			}
			if ( (sv != LRGSTAT_Both) && (sv != LRGSTAT_Null) ){
				LRGMtx->cell(mx,my)->status = sv;
				continue;
			}
		}
	}
	delete []cstmp;
	return(0);
}




/*------------------------------------------------------
	Shrink single-oriented area
------------------------------------------------------*/

int shrink_area(LRGMatrix *LRGMtx){
	int	mx,my,mxs,mys;
	int	cellstat,*cstmp;
	int	i,j,skip,r;
	int	mwidth,mheight;
	mwidth  = LRGMtx->mwidth;
	mheight = LRGMtx->mheight;
	if ( 0 == (cstmp = new int[mwidth * mheight]) )  return(-1);

				/* Copy status matrix and find null region */
	for ( my=0 ; my<mheight ; my++ ){
		for ( mx=0 ; mx<mwidth ; mx++ ){
			cellstat = LRGMtx->cell(mx,my)->status;
			if ( LRGMtx->cell(mx,my)->totalcount() == 0 )  cellstat = LRGSTAT_Null;
			cstmp[mx + (my * mwidth)] = cellstat;
		}
	}

	for ( my = 0 ; my < mheight ; my++ ){
		for ( mx = 0 ; mx < mwidth ; mx++ ){
			cellstat = cstmp[mx + (my * mwidth)];
			if ( cellstat == LRGSTAT_Both )  continue;
			if ( cellstat == LRGSTAT_Null )  continue;
			r = LRGMtx->cell(mx,my)->R;
			for ( j = -r, skip = 0 ; (j <= r) && (! skip) ; j++ ){
				mys = my + j;
				if ( mys <  0       )  continue;
				if ( mys >= mheight )  continue;
				mys *= mwidth;
				for ( i = -r ; i <= r ; i++ ){
					mxs = mx + i;
					if ( mxs <  0      )  continue;
					if ( mxs >= mwidth )  continue;
					if ( cstmp[mxs + mys] == LRGSTAT_Null )  continue;
					if ( cstmp[mxs + mys] != cellstat ){
						LRGMtx->cell(mx,my)->status = LRGSTAT_Both;
						skip = -1;
						break;
					}
				}
			}
		}
	}
	delete []cstmp;
	return(0);
}




/*------------------------------------------------------
	Load rectangle data file
------------------------------------------------------*/

int load_rectfile(LRGMatrix *LRGMtx,int ori,char *fname,int rectmode){
	struct d8pac	d8p;
	D8LOAD		D8LD;
	int		retcode;
	int		errflag;
	int		crn,rects;
	int		mx,my;
	retcode = D8LD.open(fname);
	if ( retcode < 0 )  return(retcode);
		/*  -3 : nonsupported,  -4 : memoryover,  otherwise : readerror  */

	D8LD.rewind();
	rects = 0;
	errflag = 0;
	if ( ! rectmode ){
		while ( 0 == D8LD.getdata(&d8p) ){
			if ( ori == 0 ){
				mx = (d8p.d8_data[0] + ((d8p.d8_data[2] -1) /2)) / CellSize;
				my = (d8p.d8_data[1] + ((d8p.d8_data[3] -1) /2)) / CellSize;
			}
			else{
				my = (d8p.d8_data[0] + ((d8p.d8_data[2] -1) /2)) / CellSize;
				mx = (d8p.d8_data[1] + ((d8p.d8_data[3] -1) /2)) / CellSize;
			}
			if ( LRGMtx->mwidth <= mx || LRGMtx->mheight <= my ){
				continue;
 			}
			crn = LRGMtx->cell(mx,my)->addrect(ori,
				d8p.d8_data[0],d8p.d8_data[1],d8p.d8_data[2],d8p.d8_data[3]);
			if ( crn < 0 ){
				errflag = -1;
				break;
			}
			if ( d8p.d8_data[3] > RectHeightThin )
				LRGMtx->cell(mx,my)->setattr(ori,crn,0);
			else	LRGMtx->cell(mx,my)->setattr(ori,crn,RECTSTAT_Thin);
			++rects;
		}
	}
	else{	while ( 0 == D8LD.getdata(&d8p) ){
			mx = (d8p.d8_data[0] + ((d8p.d8_data[2] -1) /2)) / CellSize;
			my = (d8p.d8_data[1] + ((d8p.d8_data[3] -1) /2)) / CellSize;
			if ( LRGMtx->mwidth <= mx || LRGMtx->mheight <= my ){
				continue;
 			}

			crn = LRGMtx->cell(mx,my)->addrect(0, \
				d8p.d8_data[0],d8p.d8_data[1], \
				d8p.d8_data[2],d8p.d8_data[3]);
			if ( crn < 0 ){ errflag = -1; break; }
			if ( d8p.d8_data[3] > RectHeightThin )
				LRGMtx->cell(mx,my)->setattr(0,crn,0);
			else	LRGMtx->cell(mx,my)->setattr(0,crn,RECTSTAT_Thin);

			crn = LRGMtx->cell(mx,my)->addrect(1, \
				d8p.d8_data[1],d8p.d8_data[0], \
				d8p.d8_data[3],d8p.d8_data[2]);
			if ( crn < 0 ){ errflag = -1; break; }
			if ( d8p.d8_data[2] > RectHeightThin )
				LRGMtx->cell(mx,my)->setattr(1,crn,0);
			else	LRGMtx->cell(mx,my)->setattr(1,crn,RECTSTAT_Thin);
			++rects;
		}
	}
	D8LD.close();
	if ( errflag )  return(-1);
	return(rects);
}




/*------------------------------------------------------
	Save rectangle data file
------------------------------------------------------*/

int save_rectfile(LRGMatrix *LRGMtx,int ori,char *fname){
	struct d4pac	d4p;
	struct d4file	d4info;
	D4SAVE		D4SV;
	LRGrp		*LRG;
	int		retcode;
	int		errflag;
	int		crn,rects,savedrects;
	int		mx,my;
	int		width,height;
	int		attr;
	CRect		*rect;
	width  = LRGMtx->mwidth  * CellSize;
	height = LRGMtx->mheight * CellSize;
	d4info.d4_count = 0;
	d4info.d4_llimit[0] = d4info.d4_llimit[1] = 0;
	d4info.d4_llimit[2] = d4info.d4_llimit[3] = 1;
	d4info.d4_hlimit[2] = width;
	d4info.d4_hlimit[3] = height;
	if ( ori == 0 ){
		d4info.d4_hlimit[0] = width -1;
		d4info.d4_hlimit[1] = height -1;
	}
	else{	d4info.d4_hlimit[1] = width -1;
		d4info.d4_hlimit[0] = height -1;
	}
	retcode = D4SV.filecreat(fname,&d4info);
	if ( 0 > retcode )  return(retcode);
		/*  -3 : nonsupported,  -4 : memoryover,  otherwise : writeerror  */

	savedrects = 0;
	errflag = 0;
	for ( my = 0 ; my < LRGMtx->mheight && (! errflag) ; my++ ){
		for ( mx = 0 ; mx < LRGMtx->mwidth ; mx++ ){
			LRG = LRGMtx->cell(mx,my);
			if ( ori == 0 ){
				if ( 0 == (LRG->status & LRGSTAT_Hrz) )  continue;
			}
			else{
				if ( 0 == (LRG->status & LRGSTAT_Vrt) )  continue;
			}
			rects = LRG->count(ori);
			for ( crn = 0 ; crn < rects ; crn++ ){
				if ( ! LRG->getrects(ori)->isactive(crn) )  continue;
				attr = LRG->getattr(ori,crn);
//				if ( attr != 0 )  continue;
				rect = LRG->getrects(ori)->getrect(crn);
				d4p.d4_data[0] = (short)rect->x1;
				d4p.d4_data[1] = (short)rect->y1;
				d4p.d4_data[2] = (short)rect->width();
				d4p.d4_data[3] = (short)rect->height();
				if ( D4SV.putdata(&d4p) ){
					errflag = -1;
					break;
				}
				++savedrects;
			}
		}
	}
	if ( 0 != D4SV.fileclose() )  errflag = -1;
	if ( errflag ){
		unlink(fname);
		return(-1);
	}
	return(savedrects);
}




/*------------------------------------------------------
	Message class
------------------------------------------------------*/

class MSG {
    public:
	void		usage(void);
	void		switcherr(void);
	void		readerr(char *name);
	void		writeerr(char *name);
	void		nonsupported(char *name);
	void		memoryover(void);
};
    	
    	
void MSG :: usage(void){		// Display Usage
	puts("txtori  v1.32, Copyright (C) 1995-2003 Hideaki Goto ");
	puts("usage : txtori [options] [-o out_file] in_file|stdin ");
	puts("		-crect    : use circumscribing rectangles (bounding boxes) ");
//	puts("		-noshrink : suppress shrinking regions ");
	puts("		-pridir   : print direction matrix on screen ");
	puts("		-ps file  : export direction matrix in postscript ");
	puts("		-quiet    : suppress messages ");
}


void MSG :: switcherr(void){
	puts("Switch error.");
}


void MSG :: readerr(char *name){
	printf("Can't read %s.\n",name);
}


void MSG :: writeerr(char *name){
	printf("Can't write %s.\n",name);
}


void MSG :: nonsupported(char *name){
	printf("Non-supported file %s.\n",name);
}


void MSG :: memoryover(void){
	puts("Not enough memory.");
}




/*------------------------------------------------------
	Argument / Option class
------------------------------------------------------*/

class ARGS {
    private:
	char		*fpath;
	char		fpathV[256];
	char		*ofname;
	char		*psfname;
    	int		sw_verbose;
	int		sw_crect;
	int		sw_pridir;
	int		sw_shrink;
    public:
	int		scanargv(int ac,char **av);
	int		SWverbose(void) {  return sw_verbose; }
	int		SWcrect(void) {  return sw_crect; }
	int		SWpridir(void) {  return sw_pridir; }
	int		SWshrink(void) {  return sw_shrink; }
	char		*infile(void);
	char		*infileV(void);
	char		*outfile(void);
	char		*outfileV(void);
	char		*psfile(void) {  return psfname; }
			ARGS(void);
};


ARGS :: ARGS(void){
	sw_verbose = -1;
	sw_crect = 0;
	sw_pridir = 0;
	sw_shrink = -1;
	fpath = 0;
	ofname = psfname = 0;
}


int ARGS :: scanargv(int ac,char **av){
	int	i,err,skip;
	char	*sw;
	err = 0;
	skip = 0;
	for( i = 1 ; i < ac ; i++ ){
		if( skip ){  skip = 0;  continue;  }
		if( *( sw = av[i] ) == '-' ){
			++sw;
			if( 0 == strcmp(sw,"crect") ){
				sw_crect = -1;
				continue;
			}
			if( 0 == strcmp(sw,"pridir") ){
				sw_pridir = -1;
				continue;
			}
			if( 0 == strcmp(sw,"noshrink") ){
				sw_shrink = 0;
				continue;
			}
			if( 0 == strcmp(sw,"ps") ){
				if( i < ac ){
					psfname = av[i+1];
					skip = -1;
					continue;
				}
				else{   err = -1;  break;  }
			}
			if( 0 == strcmp(sw,"o") ){
				if( i < ac ){
					ofname = av[i+1];
					skip = -1;
					continue;
				}
				else{   err = -1;  break;  }
			}
			if( 0 == strcmp(sw,"quiet") ){
				sw_verbose = 0;
				continue;
			}
			if( 0 == strcmp(sw,"verbose") ){
				sw_verbose = -1;
				continue;
			}
		 	err = -1;  break;
		}
		else{	fpath = av[i];
		}
	}
	return(err);
}


char * ARGS :: infile(void){
	if ( ! fpath )  return( "" );
	return( fpath );
}


char * ARGS :: infileV(void){
	if ( ! fpath )  return( "" );
	strncpy(fpathV,fpath,250);
	strcat(fpathV,"V");
	return( fpathV );
}


char * ARGS :: outfile(void){
	return( ofname );
}


char * ARGS :: outfileV(void){
	if ( 0 == ofname )  return(0);
	strncpy(fpathV,ofname,250);
	strcat(fpathV,"V");
	return( fpathV );
}




/*------------------------------------------------------
	Main routine
------------------------------------------------------*/


int main(int ac,char **av){
	ARGS		ARGS;
	MSG		MSG;
	D8LOAD		D8LD;
	PSEXP		PSEXP;
	LRGMatrix	LRGMtx;
	struct d8pac	d8p;
	int		width,height;
	int		mwidth,mheight,msize;
	int		retcode,errflag;
	int		mx,my;
	int		boxes,boxesV;

	if ( ac < 2 ){
		MSG.usage();
		exit(0);
	}
	if ( 0 != ARGS.scanargv(ac,av) ){
		MSG.switcherr();
		exit(1);
	}

	errflag = 0;

//--------	Load rectangle data files

	retcode = D8LD.open(ARGS.infile());
	switch ( retcode ){
	  case 0:	break;
	  case -3:	MSG.nonsupported(ARGS.infile());  exit(2);
	  case -4:	MSG.memoryover();  exit(3);
	  default:	MSG.readerr(ARGS.infile());  exit(2);
	}
	D8LD.getupper(&d8p);
	width  = d8p.d8_data[0] + 1;
	height = d8p.d8_data[1] + 1;
	if ( ARGS.SWverbose() )  fprintf(stderr,"Image size = %dx%d\n",width,height);
	D8LD.close();

	mwidth  = (width  + CellSize -1) / CellSize;
	width   = mwidth  * CellSize;
	mheight = (height + CellSize -1) / CellSize;
	height  = mheight * CellSize;
	msize = mwidth * mheight;

	if ( -1 == LRGMtx.alloc(mwidth,mheight,2) ){
		MSG.memoryover();  exit(3);
	}
	if ( 0 != LRGMtx.initallcells() ){
		MSG.memoryover();  exit(3);
	}

	if ( ! ARGS.SWcrect() ){
		retcode = load_rectfile(&LRGMtx,0,ARGS.infile(),0);
		if ( 0 > retcode ){
			switch ( retcode ){
			  case -3:	MSG.nonsupported(ARGS.infile());  exit(2);
			  case -4:	MSG.memoryover();  exit(3);
			  default:	MSG.readerr(ARGS.infile());  exit(2);
			}
		}
		boxes = retcode;

		retcode = load_rectfile(&LRGMtx,1,ARGS.infileV(),0);
		if ( 0 > retcode ){
			switch ( retcode ){
			  case -3:	MSG.nonsupported(ARGS.infileV());  exit(2);
			  case -4:	MSG.memoryover();  exit(3);
			  default:	MSG.readerr(ARGS.infileV());  exit(2);
			}
		}
		boxesV = retcode;
	}
	else{	retcode = load_rectfile(&LRGMtx,0,ARGS.infile(),1);
		if ( 0 > retcode ){
			switch ( retcode ){
			  case -3:	MSG.nonsupported(ARGS.infile());  exit(2);
			  case -4:	MSG.memoryover();  exit(3);
			  default:	MSG.readerr(ARGS.infile());  exit(2);
			}
		}
		boxes = boxesV = retcode;
	}

	if ( ARGS.SWverbose() ){
		fprintf(stderr,"Total %d boxes. (H)\n",boxes);
		fprintf(stderr,"Total %d boxes. (V)\n",boxesV);
	}

//--------

	LRGMtx.setradius(LocalRects);

#if 0
	for ( mx=0 ; mx<mwidth ; mx++ ){
		printf("%3d ",LRGMtx.getrects(0,mx,0,1));
	}
	printf("\n");
	for ( mx=0 ; mx<mwidth ; mx++ ){
		printf("%3d ",LRGMtx.getactrects(0,mx,0,1));
	}
	printf("\n");
#endif

#if 0
	for ( my=0 ; my<mheight ; my++ ){
		printf("%3d: ",my);
		for ( mx=0 ; mx<mwidth ; mx++ ){
			printf("%3d ",LRGMtx.cell(mx,my)->count(0));
		}
		printf("\n");
		printf("     ");
		for ( mx=0 ; mx<mwidth ; mx++ ){
			printf("%3d ",LRGMtx.cell(mx,my)->count(1));
		}
		printf("\n\n");
	}
#endif

	for ( my=0 ; my<mheight ; my++ ){
		for ( mx=0 ; mx<mwidth ; mx++ ){
			LRGMtx.cell(mx,my)->status = \
				LRGMtx.findorientation(mx,my,ARGS.SWcrect(),0);
		}
	}

	if ( ARGS.SWshrink() ){
		NR_area(&LRGMtx);
		shrink_area(&LRGMtx);
	}

	if ( ARGS.SWpridir() )	print_direction(&LRGMtx);
	if ( ARGS.psfile() )	export_direction(ARGS.psfile(),&LRGMtx);

	if ( ARGS.outfile() ){
		retcode = save_rectfile(&LRGMtx,0,ARGS.outfile());
		if ( 0 > retcode ){
			switch ( retcode ){
			  case -3:	MSG.nonsupported(ARGS.outfile());  exit(2);
			  case -4:	MSG.memoryover();  exit(3);
			  default:	MSG.writeerr(ARGS.outfile());  exit(2);
			}
		}
		if ( ARGS.SWverbose() ){
			fprintf(stderr,"Reduced to %d boxes. (H)\n",retcode);
		}
		retcode = save_rectfile(&LRGMtx,1,ARGS.outfileV());
		if ( 0 > retcode ){
			switch ( retcode ){
			  case -3:	MSG.nonsupported(ARGS.outfileV());  exit(2);
			  case -4:	MSG.memoryover();  exit(3);
			  default:	MSG.writeerr(ARGS.outfileV());  exit(2);
			}
		}
		if ( ARGS.SWverbose() ){
			fprintf(stderr,"Reduced to %d boxes. (V)\n",retcode);
		}
	}

	return(0);
}


